#include "tcp_connection.hpp"
// #include "io_thread.hpp"
#include "coroutine_hook.hpp"
#include "fd_event.hpp"
#include "io_thread.hpp"
#include "logger.hpp"
#include "reactor.hpp"
#include <cerrno>
#include <cstring>
#include <pthread.h>


namespace tinyrpc {
    TcpConnection::TcpConnection(int fd, Reactor* reactor)  : 
        m_fdEvent(fd), 
        m_mainCoroutine(std::bind(&TcpConnection::mainLoopFun, this)),
        m_reactor(reactor)
    {
        logger();
        Reactor::Task task = [this] {
            logger() << "conn coroutine is resume";
            m_mainCoroutine.resume();
        };

        reactor->addTask(task, true);

    }

    void TcpConnection::mainLoopFun() {
        while(m_state == State::Connected) {
            // TODO
            input();
            process();
            output();
        }
        logger() << "this conn loop has already break";
    }

    void TcpConnection::input() { // 调用 read_hook 读数据到应用层缓冲区
        logger() << "input";
        if(m_state == State::Disconnected) {
            logger() << "input: this conn has already break";
            return;
        }


        while(true) {
            if(m_readBuffer.getWriteable() == 0) {
                m_readBuffer.dilatation();
            }
            

            int ret = read_hook(m_fdEvent.getFd(), m_readBuffer.getWriteAddress(), m_readBuffer.getWriteable());
            if(ret == -1) {
                logger() << "read_hook ret -1 err:" << strerror(errno); 
                if(errno != EAGAIN) {
                    break;
                }
            } else if(ret == 0) { // 对端关闭了连接
                m_state = State::Disconnected;
                m_reactor->delFdEvent(&m_fdEvent);
                close(m_fdEvent.getFd());
                // IOThread::getThisIoThread()->removeFd(m_fdEvent.getFd());
                break;
            } else {
                int writeable = m_readBuffer.getWriteable();
                m_readBuffer.writeOffset(ret);
                logger() << "input_hook ret: " << ret;

                if(ret < writeable) { // 读完了结束循环
                    break;
                }

            }

        }


    }
    void TcpConnection::output() {
        logger() << "output";
        if(m_state == State::Disconnected) {
            return;
        }

        while(true) {
            if(m_writeBuffer.getReadable() == 0) {
                logger() << "no data need send";
                break;
            }

            int ret = write_hook(m_fdEvent.getFd(), m_writeBuffer.getReadAddress(), m_writeBuffer.getReadable());
            if(ret == -1) {
                logger() << "read_hook ret -1 err:" << strerror(errno); 
                break;
                
            } else if(ret == 0) { 
                logger() << "write_hook ret 0"; 
                break;
            } else {
                m_writeBuffer.readOffset(ret);
            }
            logger() << "write_hook ret: " << ret;
        }


    }
    void TcpConnection::process() {
        
        std::memcpy(m_writeBuffer.getWriteAddress(), m_readBuffer.getReadAddress(), m_readBuffer.getReadable());
        m_writeBuffer.writeOffset(m_readBuffer.getReadable());
        m_readBuffer.readOffset(m_readBuffer.getReadable());
        
        logger() << "write data " << m_writeBuffer.getReadable()  << " byte";
    }
    TcpConnection::~TcpConnection() {
        m_state = State::Disconnected;
        
    }

}